﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2017 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System.Collections.Generic;
using HeuristicLab.Common;

namespace HeuristicLab.Algorithms.DataAnalysis {
  internal sealed class SplitInfo : IDeepCloneable {
    #region Properties
    public int Position { get; private set; } // position of maximum impurity reduction
    public double MaxImpurity { get; private set; } // maximum impurity reduction
    public string SplitAttr { get; private set; } // spliting attribute
    public double SplitValue { get; private set; } // splitting value
    #endregion

    #region HLConstructors
    private SplitInfo(SplitInfo original) {
      Position = original.Position;
      MaxImpurity = original.MaxImpurity;
      SplitValue = original.SplitValue;
      SplitAttr = original.SplitAttr;
    }
    public SplitInfo() {
      MaxImpurity = -1E20;
    }
    public IDeepCloneable Clone(Cloner cloner) {
      return new SplitInfo(this);
    }
    public object Clone() {
      return new Cloner().Clone(this);
    }
    #endregion

    public void AttributeSplit(string attr, IReadOnlyList<double> attribute, IReadOnlyList<double> targets, M5CreationParameters m5CreationParams) {
      Position = 0;
      MaxImpurity = -1E20;
      SplitAttr = attr;
      SplitValue = 0.0;

      //minSize and order are fixed to 4 and 5 in weka
      //var order = Math.Max(0, minSize) + 1;
      //var len = Math.Max(1, targets.Count / order);
      //len = Math.Max(len, minSize);
      //var order = 5;

      if (targets.Count < m5CreationParams.MinLeafSize) return;

      var len = m5CreationParams.MinLeafSize;
      if (targets.Count <= m5CreationParams.MinLeafSize) return;
      var imp = m5CreationParams.Impurity.GenerateCalculator(len, targets);

      if (imp.Impurity > MaxImpurity) {
        MaxImpurity = imp.Impurity;
        SplitValue = (attribute[len - 1] + attribute[len]) / 2;
        Position = len;
      }
      for (var i = len; i < targets.Count - len; i++) {
        imp.Increment(targets[i], IncrementType.Left);
        if (attribute[i + 1].IsAlmost(attribute[i]) || imp.Impurity <= MaxImpurity) continue;
        MaxImpurity = imp.Impurity;
        SplitValue = (attribute[i] + attribute[i + 1]) / 2;
        Position = i + 1;
      }
    }
  }
}

