﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2017 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Problems.DataAnalysis;

namespace HeuristicLab.Algorithms.DataAnalysis {
  public abstract class PruningBase : ParameterizedNamedItem, IPruningType {
    private const string PruningStrengthParameterName = "PruningStrength";

    public IFixedValueParameter<DoubleValue> PruningStrengthParameter {
      get { return Parameters[PruningStrengthParameterName] as IFixedValueParameter<DoubleValue>; }
    }

    public double PruningStrength {
      get { return PruningStrengthParameter.Value.Value; }
    }

    #region Constructors & Cloning
    [StorableConstructor]
    protected PruningBase(bool deserializing) : base(deserializing) { }
    protected PruningBase(PruningBase original, Cloner cloner) : base(original, cloner) { }
    protected PruningBase() : base() {
      Parameters.Add(new FixedValueParameter<DoubleValue>(PruningStrengthParameterName, "The strength of the pruning. Higher values force the algorithm to create simpler models", new DoubleValue(2.0)));
    }
    #endregion

    #region IPruningType
    public abstract ILeafType<IRegressionModel> ModelType(ILeafType<IRegressionModel> leafType);
    public abstract void GenerateHoldOutSet(IReadOnlyList<int> allrows, IRandom random, out IReadOnlyList<int> training, out IReadOnlyList<int> holdout);
    internal virtual bool Prune(M5NodeModel node, M5CreationParameters m5CreationParams, IReadOnlyList<int> testRows, double globalStdDev) {
      if (testRows.Count == 0) return true;
      var vars = m5CreationParams.AllowedInputVariables.Concat(new[] {m5CreationParams.TargetVariable}).ToArray();
      var reducedData = new Dataset(vars, vars.Select(x => m5CreationParams.Data.GetDoubleValues(x, testRows).ToList()));
      var pd = new RegressionProblemData(reducedData, m5CreationParams.AllowedInputVariables, m5CreationParams.TargetVariable);
      pd.TrainingPartition.Start = pd.TrainingPartition.End = pd.TestPartition.Start = 0;
      pd.TestPartition.End = reducedData.Rows;

      var rmsModel = node.NodeModel.CreateRegressionSolution(pd).TestRootMeanSquaredError;
      var rmsSubTree = node.CreateRegressionSolution(pd).TestRootMeanSquaredError;

      var adjustedRmsModel = rmsModel * PruningFactor(pd.Dataset.Rows, node.NodeModelParams);
      var adjustedRmsTree = rmsSubTree * PruningFactor(pd.Dataset.Rows, node.Left.NumParam + node.Right.NumParam + 1);
      return adjustedRmsModel <= adjustedRmsTree || adjustedRmsModel < globalStdDev * 0.0001;
    }
    #endregion

    private double PruningFactor(int noInstances, int noParams) {
      return noInstances <= noParams ? 10.0 : (noInstances + PruningStrength * noParams) / (noInstances - noParams);
    }
  }
}